from bottlenest import NestFactory
from bottlenest.transports.http import HttpTransport
from .app.AppModule import AppModule


def main():
    app = NestFactory.createMicroservice(
        AppModule,
        transport=HttpTransport(port=3000),
    )

    app.listen()


if __name__ == '__main__':
    main()



import eventlet
import socketio


class WebsocketsTransportClient:
    def __init__(self):
        pass

    def setupTransport(self, context):
        self.context = context
        self.module = self.context.get('module')
        self.logger = self.context.get('logger')
        sio = socketio.Server()
        app = socketio.WSGIApp(sio)
        context.set('sio', sio)
        context.set('app', app)

    def listen(self, callback):
        # TODO: review this port to allow multiple ports
        port = self.context.get('port')
        app = self.context.get('app')
        eventlet.wsgi.server(eventlet.listen(('', port)), app)
        callback()



from bottlenest.core.NestTransport import NestTransport


class CliTransport(NestTransport):
    # called by user, passing any options
    def __init__(self):
        pass

    def getTransportKey(self):
        return 'CliTransport', f"app"

    def setupTransport(self, appContext, moduleContext):
        self.appContext = appContext
        self.moduleContext = moduleContext

    def listen(self, pool):
        pass



import eventlet
from bottlenest.core.NestTransport import NestTransport
from bottlenest.transports.http.HttpTransport import HttpTransport


class NestApplicationContext:
    # {"TransportClass":{"transportId": NestTransport}}
    transports: dict = {}
    _defaultHttpTransport: HttpTransport | None = None

    def __init__(self, moduleClass, transport=None):
        if transport is None:
            transport = HttpTransport()
        self.addTransport(transport)
        self.module = moduleClass(self)

    def addTransport(self, transport) -> None:
        group, key = transport.getTransportKey()
        if group not in NestApplicationContext.transports:
            NestApplicationContext.transports[group] = {}
        NestApplicationContext.transports[group][key] = transport
        # add to _defaultHttpTransport if needed
        print(f"NestApplicationContext.addTransport {group} {key}")
        if NestApplicationContext._defaultHttpTransport is None and group == 'HttpTransport':
            NestApplicationContext._defaultHttpTransport = transport

    def getTransport(self, transport) -> NestTransport:
        group, key = transport.getTransportKey()
        if group not in NestApplicationContext.transports:
            raise Exception(
                f'Transport "{group}"."{key}" not found')
        if key not in NestApplicationContext.transports[group]:
            raise Exception(
                f'Transport "{group}"."{key}" not found')
        return NestApplicationContext.transports[group][key]

    def getDefaultHttpTransport(self) -> HttpTransport:
        if NestApplicationContext._defaultHttpTransport is None:
            raise Exception('Default HttpTransport not found')
        return NestApplicationContext._defaultHttpTransport

    def listen(self):
        # print(NestApplicationContext.transports)
        # setup the transports
        for group in NestApplicationContext.transports.values():
            for transport in group.values():
                transport.setupTransport(self, self.module.moduleContext)
        # create the pool
        # poolQtt = len(WebsocketsFactory.__gateways__.values()) + 1
        # pool = eventlet.GreenPool(poolQtt)
        pool = eventlet.GreenPool()
        # listen to transports
        for group in NestApplicationContext.transports.values():
            for transport in group.values():
                transport.listen(pool)
        # listen to module (and providers)
        self.module.listen(pool)
        # wait for all
        try:
            print("NestApplicationContext running...")
            pool.waitall()
        except KeyboardInterrupt:
            print("NestApplicationContext keyboard interrupt")
            # ? pool.kill()



from inspect import signature
from abc import ABC, abstractmethod, ABCMeta, abstractproperty
from bottlenest.metaClasses.NestProviderContext import NestProviderContext


class NestProvider(ABC):
    # def __call__(cls, *args, **kwargs):
    #    instance = super(NestProvider, cls).__call__(*args, **kwargs)
    #    # instance.setup(*args, **kwargs)
    #    return instance

    # @abstractmethod
    @property
    def name(self):
        return self.cls.__name__

    @property
    @abstractmethod
    def eventName(self):
        return 'NestEvent'

    def __init__(self, cls):
        self.cls = cls
        self.isEnabled = False
        # self.name = cls.__name__

    def enableProvider(self):
        self.isEnabled = True

    def _getEventNames(self):
        eventClassName = self.eventName()
        eventNames = dir(self.classInstance)
        eventNames = [name for name in eventNames if type(
            getattr(self.classInstance, name)).__name__ == eventClassName]
        return eventNames

    # overridable
    # called from whithin the module's setup
    def setupProvider(self, module, moduleContext):
        print("NestProvider setupProvider")
        self._setupProvider(module, moduleContext)

    def _setupProvider(self, module, moduleContext):
        print("NestProvider _setupProvider")
        # self.module = module
        # self.moduleContext = moduleContext
        # get number of arguments in __init__
        # if self.cls has a __init__ method with 2 args
        nargs = len(signature(self.cls.__init__).parameters)
        if nargs == 2:
            eventContext = NestProviderContext(self, moduleContext)
            self.classInstance = self.cls(eventContext)
        else:
            self.classInstance = self.cls()
        moduleContext.set(self.name, self)
        # self.classInstance = self.cls(moduleContext)
        eventNames = self._getEventNames()
        for eventName in eventNames:
            # print("---->> eventName: ", eventName)
            event = getattr(self.classInstance, eventName)
            event.setupMethodDecorator(self.classInstance, moduleContext)

    def listen(self, pool):
        pass



from bottlenest.core.NestProvider import NestProvider
from .NestRoute import NestRoute
from flask import request

# TODO: add support for route prefix


class NestController(NestProvider):
    __name__ = 'NestController'

    def __init__(self, providerClass, moduleContext):
        print(f"NestController init")
        self.moduleContext = moduleContext
        # TODO: I should pass moduleContext here (conditionally)
        self.providerClass = providerClass
        self.provider = providerClass()
        # moduleContext.getOrCreateProvider(self.__name__, self)
        # transport = moduleContext.getOrCreateTransport(self.transport)
        # instance = providerClass(moduleContext)
        # for route in self.routes:
        #     route = NestRoute(route)
        #     transport.addRoute(route)

    def _getMethodDecorators(self, provider):
        for key in dir(provider):
            if type(getattr(provider, key)).__name__ == 'NestRoute':
                yield getattr(provider, key)

    def listen(self, pool):
        print(f"NestController listen")
        flaskApp = self.moduleContext.getDefaultHttpTransport().getFlaskApp()
        for route in self._getMethodDecorators(self.provider):
            route.setupMethodDecorator(self.provider, NestRequest(flaskApp))
        # transport = self.moduleContext.getOrCreateTransport(self.transport)
        # transport.listen(pool, self.callback)


##################################################################


class NestRequest:
    __name__ = 'NestRequest'

    def __init__(self, flaskApp):
        self.flaskApp = flaskApp
        self.params = NestRequestParams(request)
        self.query = {}
        self.headers = {}

    @property
    def body(self):
        try:
            return request.json
        except RuntimeError:
            # occurs when we try to get body outside a request
            pass
        return {}


class NestRequestParams(object):
    __name__ = 'NestRequestParams'

    def __init__(self, request):
        super(NestRequestParams, self).__init__()
        self.request = request

    def __getattribute__(self, __name: str):
        if __name == 'request':
            return super(NestRequestParams, self).__getattribute__(__name)
        else:
            return self.request.view_args[__name]
            # return self.request.args.get(__name, type=str)



import inquirer
import argparse
from bottlenest.core.NestProvider import NestProvider
from bottlenest.transports.cli.decorators.NestCommandArgument import NestCommandArgument
import sys


class NestCommand(NestProvider):
    __name__ = 'NestCommand'

    def __init__(self, commandClass, moduleContext, commandName, description):
        self.commandName = commandName
        self.description = description
        self.providerClass = commandClass
        self.providerName = commandClass.__name__
        self.moduleContext = moduleContext
        # TODO: send moduleContext to gatewayClass(moduleContext)
        # conditionally
        try:
            self.provider = self.providerClass()
        except TypeError:
            self.provider = self.providerClass(moduleContext)

    def parseArguments(self, parser):
        # Adding this extra command because we are proxying
        # the command that was run
        parser.add_argument(
            'commandName',
            help="The name of the command to run",
        )
        # parsing extra params
        arguments = [ag for ag in dir(self.providerClass) if isinstance(
            getattr(self.providerClass, ag), NestCommandArgument)]
        for argumentName in arguments:
            argument = getattr(self.providerClass, argumentName)
            # print(f"argument: {argumentName} {argument}")
            if argument.optional is False:
                parser.add_argument(
                    argument.argumentName,
                    **argument.kwargs,
                    required=True,
                )
            parser.add_argument(
                argument.argumentName,
                **argument.kwargs,
            )

    # def setupProvider(self, module, context):
    #     self._setupProvider(module, context)
    #     # args = sys.argv[1:]
    #     # self.provider.run(inquirer, args)

    def listen(self, pool):
        rawCommandName = sys.argv[1]
        # only run if this is the command being run
        if rawCommandName != self.commandName:
            return

        print(f"NestCommand running command: {self.commandName}")

        def _startCommandServer():
            parser = argparse.ArgumentParser(
                description=self.description,
            )
            self.parseArguments(parser)
            args = parser.parse_args()
            self.provider.run(inquirer, args)
        pool.spawn(_startCommandServer)



class InputTransportsFactory:
    """Factory for InputTransports."""

    def __init__(self, config):
        self.config = config

    def create(self, transport_name):
        """Create a new InputTransport."""
        transport_class = self.config.get('input_transports', transport_name)
        transport = transport_class()
        return transport



from bottlenest.core.NestModuleContext import NestModuleContext


class NestModule:
    def __init__(
        self,
        # from decorator
        moduleClass,
        imports,
        providers,
        controllers,
        # from NestApplicationContext
        appContext,
    ):
        self._modules = []
        self._providers = []
        # ? I'm not creating an instance of moduleClass here
        # ? should I?
        # moduleContext is exposed
        # because of NestApplicationContext.listen
        # TODO: Inject providers into moduleContext
        # each Provider should register itself in moduleContext
        # using e.g.:
        # moduleContext.registerProvider(self)
        self.moduleContext = NestModuleContext(appContext)
        for importedModule in imports:
            self._modules.append(importedModule(appContext))
        for provider in providers:
            providerInstance = provider(self.moduleContext)
            # TODO: is this assert correct?
            # assert str(type(providerInstance)).find('NestInjectable') > -1, \
            #     f"Provider {type(providerInstance)} is not NestInjectable"
            self._providers.append(providerInstance)
            self.moduleContext.registerProvider(providerInstance)
        for controller in controllers:
            providerInstance = controller(self.moduleContext)
            self._providers.append(providerInstance)
            self.moduleContext.registerProvider(providerInstance)

    def listen(self, pool):
        for module in self._modules:
            module.listen(pool)
        for provider in self._providers:
            provider.listen(pool)



from bottlenest.transports.websockets.decorators.NestWebSocketGateway import NestWebSocketGateway
from bottlenest.transports.websockets.decorators.NestSubscribeMessage import NestSubscribeMessage

# música boa
# https://www.youtube.com/watch?v=KCwk1qeh0gM
# https://www.youtube.com/watch?v=Xj4ETM2IJY8&list=PLTeMoz1bBorZjs92TBazC_QSvGOfw-CJS&index=7


def WebSocketGateway(port=80, namespace=None):
    def wrapper(gatewayClass):
        def inner(moduleContext):
            return NestWebSocketGateway(
                gatewayClass=gatewayClass,
                port=port,
                namespace=namespace,
                moduleContext=moduleContext,
            )
        return inner
    return wrapper


def SubscribeMessage(eventName):
    def wrapper(callback):
        return NestSubscribeMessage(
            callback=callback,
            eventName=eventName,
        )
    return wrapper



from abc import ABC


class NestProvider(ABC):
    __provider__ = True

    providerName: str
    providerClass: type
    provider: object

    def __init__(self, providerClass, moduleContext):
        print(f"NestProvider init")



# from ..NestHttpModule import NestHttpModule
from bottlenest.core.NestModule import NestModule
from .NestController import NestController
from .NestRoute import NestRoute
from .NestBody import NestBody


def Module(controllers=[], providers=[], imports=[]):
    def wrapper(moduleClass):
        def inner(appContext):
            return NestModule(
                moduleClass=moduleClass,
                imports=imports,
                providers=providers,
                controllers=controllers,
                appContext=appContext,
            )
        return inner
    return wrapper


def Controller():
    def wrapper(controllerClass):
        def inner(moduleContext):
            return NestController(
                controllerClass,
                moduleContext,
            )
        return inner
    return wrapper


def Body(dto):
    def wrapper(callback):
        return NestBody(
            callback=callback,
            dto=dto,
        )
    return wrapper


def Get(path):
    def wrapper(func):
        return NestRoute(
            path=path,
            method='GET',
            callback=func,
        )
    return wrapper


def Post(path):
    def wrapper(func):
        return NestRoute(
            path=path,
            method='POST',
            callback=func,
        )
    return wrapper


def Put(path):
    def wrapper(func):
        return NestRoute(
            path=path,
            method='PUT',
            callback=func,
        )
    return wrapper


def Delete(path):
    def wrapper(func):
        return NestRoute(
            path=path,
            method='DELETE',
            callback=func,
        )
    return wrapper


def Patch(path):
    def wrapper(func):
        return NestRoute(
            path=path,
            method='PATCH',
            callback=func,
        )
    return wrapper



import eventlet
from .HttpTransport import HttpTransport
from ..websockets.factories.WebsocketsFactory import WebsocketsFactory


class NestHttpModule(object):
    def __init__(self, moduleClass, imports, controllers, providers):
        self.isEnabled = False
        # module default options
        self.imports = imports
        self.controllers = controllers
        self.providers = providers
        # module variables
        self.moduleName = moduleClass.__name__
        self.moduleInstance = moduleClass()
        print(f"NestModule init", [self.moduleName])

    def enableModule(self):
        print(f"NestHttpModule enableModule {self.moduleName}")
        self.isEnabled = True
        for module in self.imports:
            print(f"enableModule", [module.moduleName])
            module.enableModule()
        for provider in self.providers:
            print(f"enableProvider", [provider.getName()])
            provider.enableProvider()

    # called automatically
    # when you run NestFactory.createMicroservice
    # (inside NestApplicationContext)
    def setupModule(self, appContext, moduleContext, transport=None):
        print(f"NestHttpModule setupModule {self.moduleName}")
        if not self.isEnabled:
            raise Exception(f"Module not enabled: {self.moduleName}")
        WebsocketsFactory.setAppContext(appContext)
        # Calls NestInputTransport.setupTransport()
        # creates the Flask app
        self.__setupInputTransport(
            transport=transport,
            appContext=appContext,
            moduleContext=moduleContext,
        )
        # run setup on any imported modules
        for module in self.imports:
            module.setupModule(appContext, moduleContext, transport)
        # run setup on providers
        # these are the most important ones
        for provider in self.providers:
            provider.setupProvider(self, moduleContext)
        # run setup on controllers
        # this is only used for HttpTransport
        for controller in self.controllers:
            controller.setupProvider(self, moduleContext)

    def __setupInputTransport(self, transport, appContext, moduleContext):
        if transport is None:
            transport = HttpTransport()
        self.transport = transport
        moduleContext.set('transport', transport)
        transport.setupTransport(
            appContext=appContext, moduleContext=moduleContext)
        return transport

    # proxy static methods
    def __getattr__(self, name):
        return getattr(self.moduleInstance, name)

    def listen(self):
        print(f"NestHttpModule listen {self.moduleName}")
        if not self.isEnabled:
            raise Exception(f"Module not enabled: {self.moduleName}")

        # print("------------------------ 1")
        # for module in self.imports:
        #     print("------------------------ 2 ", module.moduleName)
        #     if module.isEnabled:
        #         print("------------------------ 3")
        #         for provider in module.providers:
        #             print("------------------------ 4 ", provider.getName())
        #             if provider.isEnabled:
        #                 print("------------------------ 5")
        #                 provider.listen()
        #         # module.listen()

        poolQtt = len(WebsocketsFactory.__gateways__.values()) + 1
        pool = eventlet.GreenPool(poolQtt)
        # pool = eventlet.GreenPile()
        # runningServers = []

        def __listenTransport(self, pool):
            def callback():
                # self.logger.log(f"NestApplicationContext stopped")
                print(f"NestApplicationContext running callback")
            # self.transport = self.moduleContext.get('transport')
            if isinstance(self.transport, HttpTransport):
                # listen for websockets
                WebsocketsFactory.listen(pool)
                # runningServers.extend(gateways)
            # listen for http
            self.transport.listen(pool, callback)
        __listenTransport(self, pool)

        # for result in pool:
        #     print(f"NestApplicationContext result", result)
        try:
            print("NestApplicationContext running...")
            pool.waitall()
        except KeyboardInterrupt:
            print("NestApplicationContext keyboard interrupt")
            # for server in runningServers:
            #     server.kill()
        # ---
        # for server in runningServers:
        #     print(f"First server ended! {type(server)}")
        #     if isinstance(server, eventlet.greenthread.GreenThread):
        #         print("waiiit")
        #         server.wait()
        #         continue
        #     if server is not None:
        #         server.kill()
        # print("-----------------x------------------ ", runningServers)

        # while True:
        #     endAll = False
        #     for server in runningServers:
        #         # server is a GreenThread
        #         # check if it's pending
        #         if not server.dead:
        #             continue
        #         # check if it's dead
        #         else:
        #             endAll = True
        #             break
        #     if endAll:
        #         break
        #     eventlet.sleep(0.1)
        # print("NestApplicationContext ended")



from bottlenest.core.NestProvider import NestProvider


def Injectable():
    def wrapper(originalClass):
        def inner(moduleContext):
            return NestInjectable(
                originalClass=originalClass,
                moduleContext=moduleContext,
            )
        return inner
    return wrapper


class NestInjectable(NestProvider):
    __name__ = 'NestInjectable'

    def __init__(self, originalClass, moduleContext):
        self.providerName = originalClass.__name__
        self.originalClass = originalClass
        self.providerClass = originalClass
        # self.classInstance = None
        # TODO: moduleContext should be optional here
        self.classInstance = self.originalClass(moduleContext)
        self.provider = self.classInstance
        # self.dependencies = []

    def getName(self):
        return self.providerName

    def eventName(self):
        return self.providerName

    # def enableProvider(self):
    #    pass

    def setupInjectable(self, module, container):
        print(
            f"NestInjectable setupInjectable {self.providerName} {container}")
        # self.module = module
        # self.container = container
        # TODO: This should be a singleton manager
        self.classInstance = self.originalClass(container)
        # providerName = f"{module.moduleName}.{self.providerName}"
        providerName = self.providerName
        container.set(providerName, self)

    def __repr__(self):
        return f"{self.providerName}()"

    def __str__(self):
        return f"{self.providerName}()"

    def listen(self, pool):
        pass



import os
import shutil
from pathlib import Path


class FileUtils:
    @staticmethod
    def absPath(path):
        # this function should be the abs path relative to the project root
        # basePath = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        basePath = None
        while basePath is None:
            if os.path.exists(os.path.join(os.getcwd(), 'bottlenest')):
                basePath = os.getcwd() + '/bottlenest'
            else:
                os.chdir(os.path.join(os.getcwd(), '..'))
        # basePath = Path(__file__).parent.parent.parent
        # return os.path.join(basePath, path)
        return os.path.join(basePath, path)

    @staticmethod
    def copyDir(src, dst, symlinks=False, ignore=None):
        # copy dir
        # do not use for
        shutil.copytree(src, dst, symlinks, ignore)



# from bottlenest.core.NestTransport import NestTransport
from bottlenest.core.NestProvider import NestProvider
from pprint import pprint


class NestModuleContext:
    def __init__(self, appContext):
        self.appContext = appContext
        self.container = {}

    def get(self, key, defaultValue=None):
        # print("NestModuleContext get", key)
        return self.container[key] if key in self.container else defaultValue

    def set(self, key, value):
        # print("NestModuleContext set", key, value)
        self.container[key] = value

    def registerProvider(self, provider):
        # print("NestModuleContext registerProvider", provider.__name__)
        self.container[provider.__name__] = provider
        self._populateProvider(provider)

    def _populateProvider(self, provider):
        assert issubclass(type(
            provider), NestProvider), f"Provider {provider} must be a subclass of NestProvider"
        for annotation in provider.providerClass.__annotations__:
            # TODO: add providerInstance to ??
            providerInstance = provider.providerClass.__annotations__[
                annotation](self)
            setattr(provider.provider, annotation,
                    providerInstance.classInstance)

    def getOrCreateTransport(self, transport):
        try:
            transport = self.appContext.getTransport(transport)
        except Exception:
            self.appContext.addTransport(transport)
        return transport

    def getDefaultHttpTransport(self):
        return self.appContext.getDefaultHttpTransport()



from flask import request
from functools import wraps
from bottlenest.core.NestMethodDecorator import NestMethodDecorator
import json


class NestRoute(NestMethodDecorator):
    __name__ = 'NestRoute'

    def __init__(self, path, method, callback):
        self.path = path
        self.method = method
        self.callback = callback

    def setupMethodDecorator(self, provider, request):
        print(f"[NestRoute] setupMethodDecorator {self.method} {self.path}")
        request.flaskApp.add_url_rule(
            self._nestjsToFlaskPath(self.path),
            methods=[self.method],
            view_func=self._callbackWrapper(provider, self.callback, request),
        )

    def _callbackWrapper(self, provider, callback, request):
        @wraps(callback)
        def wrapped(*args, **kwargs):
            def _asJson(resp):
                if hasattr(resp, '__fields__'):
                    return resp.json()
                return json.dumps(resp)
            if issubclass(type(callback), NestMethodDecorator):
                try:
                    return _asJson(callback.setupMethodDecorator(provider, request))
                except Exception as e:
                    raise e
            return _asJson(callback(provider, request))
            # return self.callback(context)
        return wrapped

    # Flask routing is different than NestJS routing
    def _nestjsToFlaskPath(self, path: str) -> str:
        result = ''
        parts = path.split('/')
        for part in parts:
            if part.startswith(':'):
                result += '/<' + part[1:] + '>'
            else:
                result += '/' + part
        return result[1:]



from .Injectable import *
from bottlenest.transports.http.decorators import Module, Controller, Get, Post, Put, Delete, Patch, Body

# auto import from http
# from bottlenest.transports.http.errors import *
# from bottlenest.transports.http import *



import eventlet
import socketio
from bottlenest.core.NestProvider import NestProvider

servers = []


class NestWebSocketGateway(NestProvider):
    __name__ = 'NestWebSocketGateway'

    def __init__(self, gatewayClass, moduleContext, port=4001, namespace=None):
        self.port = port
        self.namespace = namespace
        self.providerClass = gatewayClass
        self.providerName = gatewayClass.__name__
        self.moduleContext = moduleContext
        # TODO: send moduleContext to gatewayClass(moduleContext)
        # conditionally
        try:
            self.provider = self.providerClass()
        except TypeError:
            self.provider = self.providerClass(moduleContext)

    def _getEvents(self, provider):
        for key in dir(provider):
            if type(getattr(provider, key)).__name__ == 'NestSubscribeMessage':
                yield getattr(provider, key)

    def listen(self, pool):
        print(f"NestSocketGateway listen {self.providerName}")

        def _startWebsocketsServer():
            print("starting websockets server")
            sio = socketio.Server()
            for event in self._getEvents(self.provider):
                event.setupMethodDecorator(self.provider, sio)
            app = socketio.WSGIApp(sio)
            eventlet.wsgi.server(eventlet.listen(('', self.port)), app)
        pool.spawn(_startWebsocketsServer)



from .factories.WebsocketsFactory import WebsocketsFactory
from bottlenest.core.NestTransport import NestTransport


class WebsocketsTransport(NestTransport):
    __sioids = 0

    # called by user, passing any options
    def __init__(self):
        pass

    def getTransportKey(self):
        WebsocketsTransport.__sioids += 1
        return 'WebsocketsTransport', f"sio-{WebsocketsTransport.__sioids}"

    def setupTransport(self, appContext, moduleContext):
        self.appContext = appContext
        self.moduleContext = moduleContext

    def listen(self, pool):
        pass

    # def listen(self, callback):
    #     # TODO: review this port to allow multiple ports
    #     port = self.appContext.get('port')
    #     app = self.appContext.get('app')
    #     eventlet.wsgi.server(eventlet.listen(('', port)), app)
    #     callback()



from abc import ABC, abstractmethod, ABCMeta


# this is the context that is given to a function inside a controller
# or a subscribe message function
# it's not the class, but the method inside a class
class NestProviderContext(ABC):
    # nestProvider is an implementation of NestProvider
    #
    def __init__(self, nestProvider, moduleContext):
        self.provider = nestProvider
        self.moduleContext = moduleContext

    def get(self, key):
        if key.startswith('provider.'):
            return self.moduleContext.get(key.replace('provider.', ''))
        # getName = f"{self.provider.module.moduleName}.{key}"
        # return self.moduleContext.get(getName)
        return self.moduleContext.get(key)

    def inject(self, injectable):
        # getName = f"{self.provider.module.moduleName}.{injectable.providerName}"
        # provider = self.moduleContext.get(getName)
        provider = self.moduleContext.get(injectable.providerName)
        if provider is None:
            raise Exception(f"Provider not found: {injectable.providerName}")
        return provider.classInstance



from bottlenest.core.NestMethodDecorator import NestMethodDecorator
from pydantic.error_wrappers import ValidationError
from bottlenest.transports.http.errors.BadRequestError import BadRequestError


class NestBody(NestMethodDecorator):
    __name__ = 'NestBody'

    def __init__(self, callback, dto):
        self.callback = callback
        self.dto = dto

    def setupMethodDecorator(self, moduleContext, request):
        # pydantic validates automatically
        self.dto(request.body)
        return self.callback(moduleContext, request)



from bottlenest.transports.cli.decorators.NestCommand import NestCommand
from bottlenest.transports.cli.decorators.NestCommandArgument import NestCommandArgument


def Command(name, *args, **kwargs):
    def wrapper(commandClass):
        def inner(moduleContext):
            return NestCommand(
                commandClass=commandClass,
                commandName=name,
                moduleContext=moduleContext,
                *args,
                **kwargs,
            )
        return inner
    return wrapper


def CommandArgument(name, optional=None, *args, **kwargs):
    def wrapper(callback):
        return NestCommandArgument(
            callback=callback,
            argumentName=name,
            optional=optional,
            *args,
            **kwargs,
        )
    return wrapper



from bottlenest.core.NestApplicationContext import NestApplicationContext
import os


class NestFactory:
    @staticmethod
    def create(moduleClass, transport=None):
        return NestApplicationContext(
            moduleClass=moduleClass,
            transport=transport,
        )

    @staticmethod
    def createMicroservice(moduleClass, transport=None):
        return NestApplicationContext(
            moduleClass=moduleClass,
            transport=transport,
        )



from bottlenest.transports.cli import Command, CommandArgument
from bottlenest.utils.FileUtils import FileUtils


@Command(
    name="new",
    description="Create a new project",
)
class NewCoreCommand:
    __name__ = 'NewCoreCommand'

    # context is the provider Context
    def __init__(self, context):
        self.context = context

    def run(self, context, args):
        print(
            f"Creating project '{args.name}' from template '{args.template}'...")
        exampleFolder = f"templates/{args.template}"
        # copy exampleFolder to current folder
        FileUtils.copyDir(FileUtils.absPath(exampleFolder), f"./{args.name}")
        print(f"Project '{args.name}' created!")

    @CommandArgument(
        name="name",
        help="Project name",
        optional=True,
    )
    def nameArg(self, value):
        return value

    @CommandArgument(
        name="--template",
        dest="template",
        help="The template to use",
        default="helloworld",
        optional=True,
    )
    def templateArg(self, value):
        return value



from bottlenest.common.Injectable import Injectable
from dotenv import load_dotenv
from bottlenest.common import Module
from bottlenest.config.ConfigService import ConfigService
import os


@Module(
    providers=[ConfigService],
)
class ConfigModule:
    pass


def forRoot(envFilePath='.env'):
    # load .env file and add to environment
    # to load a .env file we will use the dotenv package
    load_dotenv(envFilePath)
    return ConfigModule


ConfigModule.forRoot = staticmethod(forRoot)



from bottlenest.core.NestMethodDecorator import NestMethodDecorator


class NestSubscribeMessage(NestMethodDecorator):
    __name__ = 'NestSubscribeMessage'

    def __init__(self, callback, eventName):
        self.eventName = eventName
        self.callback = callback

    # called from whithin NestWebsocketGateway._setupProvider()
    def setupMethodDecorator(self, cls, sio):
        print(f"setup event {self.eventName}")

        @sio.on(self.eventName)
        def _callback(sid, data):
            print(f"NestSubscribeMessage {self.eventName}")
            result = self.callback(cls, data)
            if result is not None:
                sio.emit(self.eventName, result, room=sid)
